local BadgeService = game:GetService("BadgeService")
local Promise = require(script.Promise.Value)
local Binder = require(script.Binder.Value)

local getBadgeInfoPromise = Promise.promisify(function (...) return BadgeService:GetBadgeInfoAsync(...) end)
local awardBadgePromise = Promise.promisify(function (...) return BadgeService:AwardBadge(...) end)
local userHasBadgePromise = Promise.promisify(function (...) return BadgeService:UserHasBadgeAsync(...) end)

local BadgeAwarder = {}
BadgeAwarder.__index = BadgeAwarder
BadgeAwarder.ATTR_BADGE_ID = "BadgeId"
BadgeAwarder.TAG = "BadgeAwarder"

BadgeAwarder.BADGE_ALREADY_OWNED = "BadgeAlreadyOwned"
BadgeAwarder.BADGE_NOT_ENABLED = "BadgeNotEnabled"
BadgeAwarder.BADGE_AWARDED = "BadgeAwarded"
BadgeAwarder.Binder = Binder.new(BadgeAwarder.TAG, BadgeAwarder)

function BadgeAwarder.new(object)
	local self = setmetatable({
		object = assert(typeof(object) == "Instance" and object, "Instance expected");
		_conns = {};
		_awardPromises = {};
		_badgeInfo = nil;
	}, BadgeAwarder)
	
	self.touchedFunc = function (...) return self:onBasePartTouched(...) end
	self.mouseClickFunc = function (...) return self:onClickDetectorMouseClicked(...) end
	self.triggeredFunc = function (...) return self:onProximityPromptTriggered(...) end
	
	self:getBadgeId()
	self:connectRecursive(self.object)
	
	return self
end

function BadgeAwarder:destroy()
	-- Cancel pending promises (they should finally remove themselves from this table)
	for promise in pairs(self._awardPromises) do
		promise:cancel()
	end
	self._awardPromises = nil
	-- Disconnect active connections
	for object, conn in pairs(self._conns) do
		conn:Disconnect()
		self._conns[object] = nil
	end
	self._conns = nil
	-- Forget everything else
	self.object = nil
	self._badgeInfo = nil
end

function BadgeAwarder:getBadgeId()
	return assert(
		self.object:GetAttribute(BadgeAwarder.ATTR_BADGE_ID),
		("%s missing attribute %s"):format(
			self.object:GetFullName(),
			BadgeAwarder.ATTR_BADGE_ID
		)
	)
end

function BadgeAwarder:connectRecursive(object)
	self:connect(object)
	for _, child in pairs(object:GetChildren()) do
		self:connectRecursive(child)
	end
end

function BadgeAwarder:connect(object)
	local conn = nil
	if object:IsA("BasePart") then
		object.Touched:Connect(self.touchedFunc)
	elseif object:IsA("ProximityPrompt") then
		object.Triggered:Connect(self.triggeredFunc)
	elseif object:IsA("ClickDetector") then
		object.MouseClick:Connect(self.mouseClickFunc)
	end
	if conn then
		self._conns[object] = conn
	end
end

do
	local Players = game:GetService("Players")
	function BadgeAwarder:onBasePartTouched(otherPart)
		if not otherPart.Parent then return end
		local player = Players:GetPlayerFromCharacter(otherPart.Parent)
		if not player then return end
		
		self:awardBadge(player)
	end
end

function BadgeAwarder:onClickDetectorMouseClicked(player)
	self:awardBadge(player)
end

function BadgeAwarder:onProximityPromptTriggered(player)
	self:awardBadge(player)
end

function BadgeAwarder:awardBadge(player)
	assert(typeof(player) == "Instance" and player:IsA("Player"), "Player expected")
	if self._awardPromises[player] then
		return self._awardPromises[player]
	end
	local badgeId = self:getBadgeId()
	self._awardPromises[player] = userHasBadgePromise(player.UserId, badgeId):andThen(function (hasBadge)
		if not hasBadge then
			return getBadgeInfoPromise(badgeId):andThen(function (badgeInfo)
				self._badgeInfo = badgeInfo
				if badgeInfo["IsEnabled"] then
					return awardBadgePromise(player.UserId, badgeId):andThen(function ()
						return Promise.resolve(BadgeAwarder.BADGE_AWARDED)
					end)
				else
					return Promise.reject(BadgeAwarder.BADGE_NOT_ENABLED)
				end
			end)
		else
			return Promise.reject(BadgeAwarder.BADGE_ALREADY_OWNED)
		end
	end):catch(function (err)
		warn("BadgeAwarder error", self.object:GetFullName(), tostring(err))
	end):finally(function ()
		self._awardPromises[player] = nil
	end)
end

return BadgeAwarder
